from typing import List
from django.conf import settings
from django.contrib.auth.models import Permission, Group
from django.apps import apps as django_apps
from django.core.exceptions import ImproperlyConfigured


def set_profile(user, groups: List[Group], permissions: List[Permission]):
    user_groups = user.groups.all()
    user_permissions = user.get_user_permissions()
    add_groups = []
    add_permissions = []
    delete_groups = []
    delete_permissions = []
    
    for permission in user_permissions:
        if permission not in ['{}.{}'.format(p.content_type.app_label, p.codename) for p in permissions]:
            delete_permissions.append(permission)
            
    for group in user_groups:
        if group.name not in [g.name for g in groups]:
            delete_groups.append(group)
            
    for permission in permissions:
        if '{}.{}'.format(permission.content_type.app_label, permission.codename) not in [p for p in user_permissions]:
            add_permissions.append(permission)
            
    for group in groups:
        if group.name not in [g.name for g in user_groups]:
            add_groups.append(group)
    
    for group in add_groups:
        group.user_set.add(user)
        
    for permission in add_permissions:
        user.user_permissions.add(permission)

    for group in add_groups:
        group.user_set.remove(user)
        
    for permission in delete_permissions:
        data = permission.split('.')
        perm = Permission.objects.get(content_type__app_label=data[0],codename=data[1])
        perm.user_set.remove(user)
        
        
def get_user_model_keycloak():
    """
    Return the User Keycloak model that is active in this project.
    """
    try:
        return django_apps.get_model(settings.KEYCLOAK['KEYCLOAK_USER_MODEL'], require_ready=False)
    except ValueError:
        raise ImproperlyConfigured("KEYCLOAK_USER_MODEL must be of the form 'app_label.model_name'")
    except LookupError:
        raise ImproperlyConfigured(
            "KEYCLOAK_USER_MODEL refers to model '%s' that has not been installed" % settings.KEYCLOAK['KEYCLOAK_USER_MODEL']
        )


from django.contrib.auth.models import Group, Permission
from typing import List
from django.db.models import Value
from django.db.models.functions import Concat
from django.contrib.contenttypes.models import ContentType


class GroupsPermissions:
    roles = []
    permissions = []

    def __init__(self, roles: List[str]) -> None:
        self.roles = [rol.replace('_role', '') for rol in list(
            set(roles)) if rol.endswith('_role')]
        self.permissions = [
            permission.replace('_permission', '') for permission in roles if permission.endswith('_permission')]

    def get_roles(self) -> List[Group]:
        # roles_available = list(Group.objects.filter(name__in=self.roles))
        roles_available = [Group(name=rol) for rol in self.roles]
        return roles_available

    def get_permissions(self) -> List[Permission]:
        # queryset = Permission.objects.annotate(search_permission=Concat('content_type__app_label', Value('_'), 'codename'))
        # permissions_available = list(queryset.filter(search_permission__in=self.permissions))
        ct = ContentType(app_label='permission', model='account')
        permissions_available = [Permission(
            name=permission, codename=permission, content_type=ct) for permission in self.permissions]
        return permissions_available

    def get_permissions_in_user(self) -> List[Permission]:
        # permissions_avalible = list([permision for permision in self.get_permissions() if permision.pk not in [
        #     permisions.pk for list_permisions in [list(rol.permissions.all()) for rol in self.get_roles()] for permisions in list_permisions]])
        return self.get_permissions()



from django_keycloak.keycloak import get_default_master_keycloak_admin


def create_role_by_owner(role_model):
    keycloak_admin = get_default_master_keycloak_admin()
    keycloak_admin.set_realm_name(realm_name=role_model.owner.name)
    payload = {
        "id": str(role_model.id),
        "name": role_model.role.name.replace(' ', '_').lower() + '_role',
        "description": role_model.role.description,
        "clientRole": False,
        "attributes": {
            "is_editable": [role_model.role.is_editable],
            "is_role": [True]
        }
    }
    keycloak_admin.create_realm_role(payload=payload)


def create_role_by_app(role_model):
    keycloak_admin = get_default_master_keycloak_admin()
    keycloak_admin.set_realm_name(realm_name=role_model.apps_owner.owner.name)
    payload = {
        "id": str(role_model.id),
        "name": role_model.role.name.replace(' ', '_').lower() + '_role',
        "description": role_model.role.description,
        "clientRole": True,
        "attributes": {
            "is_editable": [role_model.role.is_editable],
            "is_role": [True]
        }
    }
    client_role_id = keycloak_admin.get_client_id(
        client_id=role_model.apps_owner.app.client_id)
    keycloak_admin.create_client_role(
        client_role_id=client_role_id, payload=payload)


def delete_role_by_owner(role_model):
    keycloak_admin = get_default_master_keycloak_admin()
    keycloak_admin.set_realm_name(realm_name=role_model.owner.name)
    keycloak_admin.delete_realm_role(role_name=f"{role_model.role.name.replace(' ', '_').lower()}_role")


def delete_role_by_app(role_model):
    keycloak_admin = get_default_master_keycloak_admin()
    keycloak_admin.set_realm_name(realm_name=role_model.apps_owner.owner.name)
    client_role_id = keycloak_admin.get_client_id(
        client_id=role_model.apps_owner.app.client_id)
    keycloak_admin.delete_client_role(client_role_id=client_role_id,role_name=f"{role_model.role.name.replace(' ', '_').lower()}_role")


from django.core.management.base import BaseCommand
from django.contrib.auth.models import Group, Permission
from django.db.models import Value
from django.db.models.functions import Concat
from kafka import KafkaConsumer
import json
from django.conf import settings
from django.contrib.auth import get_user_model
from kafka.consumer.fetcher import ConsumerRecord


def order_by_timestamp(record:ConsumerRecord):
    return record.timestamp

def to_boolean(value):
    if isinstance(value, str) and value.lower() in ['true', 'false']:
        return value.lower() == 'true'
    return value

class Command(BaseCommand):
    help = 'User change consumer kafka'
    running = True
    bootstrap_servers = ""
    refresh_time = 0
    mechanism = ""
    username = ""
    password = ""
    topic = ""
    group_id = ""
    attribute_fillable = []


    def add_arguments(self, parser):
        parser.add_argument('-b', '--bootstrap_servers', nargs='?', const=1, type=str,
                            default=settings.KEYCLOAK['KEYCLOAK_KAFKA_BOOTSTRAP_SERVERS'],
                            help='Define the bootstrap_servers kafka')
        parser.add_argument('-r', '--refresh_time', nargs='?', const=1, type=int,
                            default=settings.KEYCLOAK['KEYCLOAK_KAFKA_REFRESH_TIME'],
                            help='Consumer refresh time kafka')
        parser.add_argument('-m', '--mechanism', nargs='?', const=1, type=str,
                            default=settings.KEYCLOAK['KEYCLOAK_KAFKA_MECHANISM'], help='Define a mechanism kafka')
        parser.add_argument('-u', '--username', nargs='?', const=1, type=str,
                            default=settings.KEYCLOAK['KEYCLOAK_KAFKA_USERNAME'], help='Define a username kafka')
        parser.add_argument('-p', '--password', nargs='?', const=1, type=str,
                            default=settings.KEYCLOAK['KEYCLOAK_KAFKA_PASSWORD'], help='Define a password kafka')
        parser.add_argument('-t', '--topic', nargs='?', const=1, type=str,
                            default=settings.KEYCLOAK['KEYCLOAK_KAFKA_TOPIC'], help='Define a topic kafka')
        parser.add_argument('-g', '--group', nargs='?', const=1, type=str,
                            default=settings.KEYCLOAK['KEYCLOAK_KAFKA_TOPIC_GROUP'], help='Define a topic kafka')


    def handle(self, *args, **options):
        self.bootstrap_servers = options['bootstrap_servers']
        self.refresh_time = options['refresh_time']
        self.mechanism = options['mechanism']
        self.username = options['username']
        self.password = options['password']
        self.topic = options['topic']
        self.group_id = options['group']
        self.attribute_fillable = settings.KEYCLOAK.get('KEYCLOAK_KAFKA_ATRIBUTES_FILLABLE_USER', [])
        self.basic_consume_loop()


    def user_create(self, payload_data):
        payload = {}
        for config in self.attribute_fillable:
            if isinstance(config, tuple):
                if config[1].startswith('attributes'):
                    if 'attributes' in payload_data['representation'] and config[1].replace('attributes.', '') in \
                            payload_data['representation']['attributes']:
                        payload[config[0]] = \
                            to_boolean(
                                payload_data['representation']['attributes'][config[1].replace('attributes.', '')][0])
                else:
                    payload[config[0]] = to_boolean(payload_data['representation'][config[1]])
                continue
            payload[config] = payload_data['representation'][config]
        payload['id'] = payload_data['user_id']
        UserModel = get_user_model()
        user = UserModel(**payload)
        user.save()


    def client_role_mapping_create(self, payload_data):
        client_id = payload_data['client_id']
        if client_id in settings.KEYCLOAK['CLIENT_ID_VALIDS']:
            payload_data['representation'] = payload_data['representation'][0]
            UserModel = get_user_model()
            user = UserModel.objects.get(pk=payload_data['user_id'])
            name = payload_data['representation']['name']
            if name.endswith('_role'):
                group = Group.objects.get(name=name.replace('_role',''))
                group.user_set.add(user)
            else:
                queryset = Permission.objects.annotate(
                    search_permission=Concat('content_type__app_label', Value('_'), 'codename'))
                permissions_available = list(queryset.filter(search_permission=name.replace('_permission','')))
                for perm in permissions_available:
                    perm.user_set.add(user)


    def user_delete(self, payload_data):
        UserModel = get_user_model()
        try:
            user = UserModel.objects.get(pk=payload_data['user_id'])
            user.delete()
        except Exception as e:
            print('user not exist %s' % str(e))


    def client_role_mapping_delete(self, payload_data):
        client_id = payload_data['client_id']
        if client_id in settings.KEYCLOAK['CLIENT_ID_VALIDS']:
            payload_data['representation'] = payload_data['representation'][0]
            UserModel = get_user_model()
            user = UserModel.objects.get(pk=payload_data['user_id'])
            name = payload_data['representation']['name']
            if name.endswith('_role'):
                group = Group.objects.get(name=name.replace('_role', ''))
                group.user_set.remove(user)
            else:
                queryset = Permission.objects.annotate(
                    search_permission=Concat('content_type__app_label', Value('_'), 'codename'))
                permissions_available = list(queryset.filter(search_permission=name.replace('_permission', '')))
                for perm in permissions_available:
                    perm.user_set.remove(user)


    def user_update(self, payload_data):
        UserModel = get_user_model()
        user = UserModel.objects.get(pk=payload_data['user_id'])
        payload = {}
        for config in self.attribute_fillable:
            if isinstance(config, tuple):
                if config[1].startswith('attributes'):
                    if 'attributes' in payload_data['representation'] and config[1].replace('attributes.', '') in \
                            payload_data['representation']['attributes']:
                        payload[config[0]] = \
                            to_boolean(
                                payload_data['representation']['attributes'][config[1].replace('attributes.', '')][0])
                else:
                    payload[config[0]] = to_boolean(payload_data['representation'][config[1]])
                continue
            payload[config] = payload_data['representation'][config]
        for key, value in payload.items():
            setattr(user, key, value)
        user.save()


    def client_role_mapping_update(self, payload):
        print('update_role_permission no implementable')
        print(payload)


    def basic_consume_loop(self):
        consumer = None
        try:
            consumer = KafkaConsumer(self.topic,
                                     group_id=self.group_id,
                                     bootstrap_servers=self.bootstrap_servers.split(','),
                                     consumer_timeout_ms=self.refresh_time)

            while self.running:
                list_messages = [message for message in consumer]
                list_messages.sort(key=order_by_timestamp)
                call_methods = {
                    "USER_CREATE": self.user_create,
                    "CLIENT_ROLE_MAPPING_CREATE": self.client_role_mapping_create,
                    "USER_DELETE": self.user_delete,
                    "CLIENT_ROLE_MAPPING_DELETE": self.client_role_mapping_delete,
                    "USER_UPDATE": self.user_update,
                    "CLIENT_ROLE_MAPPING_UPDATE": self.client_role_mapping_update
                }
                for message in list_messages:
                    try:
                        message = json.loads(message.value.decode('utf-8'))
                        if 'resourceType' in message and 'operationType' in message and message['resourceType'] in ['USER', 'CLIENT_ROLE_MAPPING']:
                            method = "{}_{}".format(message['resourceType'], message['operationType'])
                            if 'representation' in message and isinstance(message['representation'], str):
                                message['representation'] = json.loads(message['representation'])
                                if 'attributes' in message['representation'] and isinstance(message['representation']['attributes'], str):
                                    message['representation']['attributes'] = json.loads(message['representation']['attributes'])
                            call_methods[method](message)
                            print(message)
                    except Exception as e:
                        print('error %s' % str(e))
        finally:
            # Close down consumer to commit final offsets.
            print('Close down consumer to commit final offsets.')
            if consumer:
                consumer.close()


from rest_framework import permissions
from django.core.exceptions import PermissionDenied
from typing import List
import inspect
from django.contrib.auth.models import Group, Permission

class ValidatePermissions(permissions.BasePermission):
    """
    Allows access only to authenticated users.
    """

    def _has_list_has_permissions(self, permissions_list: List, request, view):
        for per in permissions_list:
            if not (
                    per().has_permission(request, view) if (
                            callable(per) and 'has_permission' in dir(per())) else request.user.has_perm(per)):
                return False
        return True

    def _validate_permission(self, permission, request, view):
        if callable(permission) and 'has_permission' in dir(permission()):
            return permission().has_permission(request, view)

        if isinstance(permission, str):
            return request.user.has_perm(permission)

        if not permission and isinstance(permission, List):
            return True

        if permission and isinstance(permission, List):
            return self._has_list_has_permissions(permission, request, view)

        return False

    def has_permission(self, request, view):
        valid = False
        if request.user and request.user.is_authenticated and hasattr(view,
                                                                      'validate_permissions') and view.validate_permissions:
            if isinstance(view.validate_permissions, str) or isinstance(view.validate_permissions, List) or callable(
                    view.validate_permissions) and 'has_permission' in dir(view.validate_permissions()):
                permission = view.validate_permissions
            else:
                if not view.action in view.validate_permissions:
                    return True
                permission = view.validate_permissions[view.action]
            valid = self._validate_permission(permission, request, view)

        # In case the 403 handler should be called raise the exception
        if hasattr(view, 'raise_exception_validate') and view.raise_exception_validate and not valid:
            raise PermissionDenied
        return valid


def single_permission(permission: str):
    class C(permissions.BasePermission):
        def has_permission(self, request, view):
            if request.user and request.user.is_authenticated:
                return request.user.has_perm(permission)
            return False

    return C


def single_group(group: str):
    class C(permissions.BasePermission):
        def has_permission(self, request, view):
            if request.user and request.user.is_authenticated:
                return group in [g.name for g in request.user.groups.all()]
            return False

    return C


from django.contrib import admin, messages
# from keycloak.exceptions import KeycloakClientError
from requests.exceptions import HTTPError
from django_keycloak.models import Client
from django_keycloak.services.synchronize import RolePermissionKeyclock

# from django_keycloak.models import (
#     Client,
#     OpenIdConnectProfile,
#     RemoteClient
# )
# import django_keycloak.services.permissions
# import django_keycloak.services.realm
# import django_keycloak.services.uma


# def refresh_open_id_connect_well_known(modeladmin, request, queryset):
#     for realm in queryset:
#         django_keycloak.services.realm.refresh_well_known_oidc(realm=realm)
#     modeladmin.message_user(
#         request=request,
#         message='OpenID Connect .well-known refreshed',
#         level=messages.SUCCESS
#     )


# refresh_open_id_connect_well_known.short_description = 'Refresh OpenID ' \
#                                                        'Connect .well-known'


# def refresh_certs(modeladmin, request, queryset):
#     for realm in queryset:
#         django_keycloak.services.realm.refresh_certs(realm=realm)
#     modeladmin.message_user(
#         request=request,
#         message='Certificates refreshed',
#         level=messages.SUCCESS
#     )


# refresh_certs.short_description = 'Refresh Certificates'


# def clear_client_tokens(modeladmin, request, queryset):
#     OpenIdConnectProfile.objects.filter(realm__in=queryset).update(
#         access_token=None,
#         expires_before=None,
#         refresh_token=None,
#         refresh_expires_before=None
#     )
#     modeladmin.message_user(
#         request=request,
#         message='Tokens cleared',
#         level=messages.SUCCESS
#     )


# clear_client_tokens.short_description = 'Clear client tokens'


def synchronize_permissions(modeladmin, request, queryset):
    for realm in queryset:
        try:
            # django_keycloak.services.permissions.synchronize(
            #     client=realm.client)
            # print(modeladmin.name)
            rolepermissionkeyclock = RolePermissionKeyclock(realm=realm)
            rolepermissionkeyclock.synchronize_permissions_roles()
            print(realm)
        except Exception as e:
            print(str(e))
            modeladmin.message_user(
                request=request,
                message='Forbidden for {}. Does the client\'s service '
                        'account has the "keycloak_client" role?'.format(
                            realm.name
                        ),
                level=messages.ERROR
            )
    modeladmin.message_user(
        request=request,
        message='Permissions synchronized',
        level=messages.SUCCESS
    )


synchronize_permissions.short_description = 'Synchronize permissions'


def synchronize_resources(modeladmin, request, queryset):
    for realm in queryset:
        try:
            # django_keycloak.services.uma.synchronize_client(
            #     client=realm.client)
            # print(modeladmin.name)
            print(realm)
        except Exception as e:
            modeladmin.message_user(
                request=request,
                message='Forbidden for {}. Does the client\'s service '
                        'account has the "keycloak_client" role?'.format(
                            realm.name
                        ),
                level=messages.ERROR
            )
    modeladmin.message_user(
        request=request,
        message='Resources synchronized',
        level=messages.SUCCESS
    )


synchronize_resources.short_description = 'Synchronize models as Keycloak ' \
                                          'resources'


class ClientAdmin(admin.TabularInline):

    model = Client
    
    extra = 1

    fields = ('client_id', )
    


class RealmAdmin(admin.ModelAdmin):

    inlines = [ClientAdmin]

    actions = [
        synchronize_permissions,
        synchronize_resources
    ]

    fieldsets = (
        (None, {
            'fields': ('name', 'url', )
        }),
        ('Location', {
            'fields': ('client_id', 'secret',)
        }),
        ('User', {
            'fields': ('username_or_email', 'password',)
        }),

    )


# Generated by Django 3.2.16 on 2022-10-27 04:33

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='User',
            fields=[
                ('id', models.CharField(max_length=150, primary_key=True, serialize=False)),
                ('email', models.EmailField(max_length=254, verbose_name='email address')),
                ('username', models.CharField(max_length=150, verbose_name='username')),
                ('first_name', models.CharField(blank=True, max_length=150, verbose_name='first name')),
                ('last_name', models.CharField(blank=True, max_length=150, verbose_name='last name')),
            ],
            options={
                'verbose_name': 'User',
                'verbose_name_plural': 'Users',
                'db_table': 'user_keycload',
                'managed': False,
            },
        ),
        migrations.CreateModel(
            name='Realm',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('url', models.CharField(help_text='Url as known on the Keycloak server. This url is used in the API paths of this Realm.', max_length=255)),
                ('name', models.CharField(help_text='Name as known on the Keycloak server. This name is used in the API paths of this Realm.', max_length=255, unique=True)),
                ('client_id', models.CharField(help_text='Client id as known on the Keycloak server. This name is used in the API paths of this Realm.', max_length=255)),
                ('secret', models.CharField(help_text='Secret of client id on the Keycloak server.', max_length=255)),
                ('username_or_email', models.CharField(help_text='User of client id on the Keycloak server.', max_length=255)),
                ('password', models.CharField(help_text='Password of user on the Keycloak server.', max_length=255)),
            ],
            options={
                'verbose_name': 'Realm',
                'verbose_name_plural': 'Realms',
                'db_table': 'realm',
                'managed': True,
            },
        ),
        migrations.CreateModel(
            name='Client',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('client_id', models.CharField(max_length=255)),
                ('realm', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='client', to='django_keycloak.realm')),
            ],
            options={
                'verbose_name': 'Realm Client',
                'verbose_name_plural': 'Realm Clients',
                'db_table': 'realm_client',
                'managed': True,
            },
        ),
    ]



from django_keycloak.keycloak import get_default_master_keycloak_admin


def create_client_by_app_owner(client_model, realm_model, id=None):
    keycloak_admin = get_default_master_keycloak_admin()
    payload = dict(client_model.base_config)
    if id:
        payload['id'] = str(id)
    payload['clientId'] = client_model.client_id
    payload['name'] = client_model.name
    payload['rootUrl'] = client_model.rootUrl
    payload['baseUrl'] = client_model.baseUrl
    payload['enabled'] = client_model.active
    payload['publicClient'] = not client_model.is_api
    if not client_model.is_api:
        payload['redirectUris'] = client_model.redirectUris.split(",")
        payload['webOrigins'] = client_model.webOrigins.split(",")
        payload['attributes']['post.logout.redirect.uris'] = "##".join(client_model.logoutUris.split(","))
    keycloak_admin.create_client_by_realm_name(realm_name=realm_model.realm, payload=payload)
    if client_model.is_api:
        keycloak_admin.set_realm_name(realm_name=realm_model.realm)
        client_id = keycloak_admin.get_client_id(client_model.client_id)
        keycloak_admin.generate_client_secrets(client_id=client_id)


def delete_client_by_app_owner(client_model, realm_model):
    keycloak_admin = get_default_master_keycloak_admin()
    keycloak_admin.set_realm_name(realm_name=realm_model.realm)
    client_id = keycloak_admin.get_client_id(client_model.client_id)
    keycloak_admin.delete_client(client_id=client_id)


import json

from django.db.models import QuerySet

from django_keycloak.keycloak import get_default_master_keycloak_admin


def get_base_user(user_model, password):
    return {
        "id": str(user_model.id),
        "email": user_model.email,
        "emailVerified": user_model.email_verified,
        "enabled": True,
        "firstName": user_model.first_name,
        "lastName": user_model.last_name,
        "username": user_model.username,
        "attributes": {
            "is_staff": user_model.is_staff,
            "is_superuser": user_model.is_superuser
        },
        "credentials": [{
            "type": "password",
            "value": password,
            "temporary": False
        }]
    }


def create_user_by_owner(user_model, password):
    keycloak_admin = get_default_master_keycloak_admin()
    keycloak_admin.set_realm_name(realm_name=user_model.realm_name)
    payload = get_base_user(user_model, password)
    keycloak_admin.create_user(payload=payload)
    return keycloak_admin.get_user_id(user_model.username)


def create_user_by_master(user_model, password):
    keycloak_admin = get_default_master_keycloak_admin()
    payload = get_base_user(user_model, password)
    keycloak_admin.create_user(payload=payload)


def get_roles_representation(role_names, client_id, keycloak_admin):
    roles_representations = keycloak_admin.get_client_roles(client_id=client_id)
    return [r for r in roles_representations if r['name'] in role_names]


def assign_client_role_to_user(user_model, role_model):
    keycloak_admin = get_default_master_keycloak_admin()
    keycloak_admin.set_realm_name(realm_name=user_model.realm_name)

    if not isinstance(role_model, QuerySet):
        roles = [f"{role_model.role.name.replace(' ', '_').lower()}_role"]
    else:
        roles = [f"{r.role.name.replace(' ', '_').lower()}_role" for r in role_model]

    client_id = keycloak_admin.get_client_id(
        client_id=role_model[0].apps_owner.app.client_id)
    roles = get_roles_representation(role_names=roles,client_id=client_id,keycloak_admin=keycloak_admin)
    keycloak_admin.assign_client_role(user_id=str(user_model.id), client_id=client_id, roles=roles)


def get_base_business_unit(business_units_model):
    business_unit_dict= {
        "id" : business_units_model.id,
        "name": business_units_model.name
    }
    return json.dumps(business_unit_dict)


def get_attributes(user_model):
    keycloak_admin = get_default_master_keycloak_admin()
    keycloak_admin.set_realm_name(realm_name=user_model.realm_name)
    user_representation = keycloak_admin.get_user(user_id=str(user_model.id))
    if not 'attributes' in user_representation:
        return {}
    return user_representation['attributes']


def assign_business_units_to_user(user_model, business_units):
    keycloak_admin = get_default_master_keycloak_admin()
    keycloak_admin.set_realm_name(realm_name=user_model.realm_name)
    attributes = get_attributes(user_model)
    # if not 'business_units' in attributes:
    #     attributes['business_units'] = []
    attributes['business_units'] = business_units
    payload = {
        "attributes": {
            **attributes
        }
    }
    keycloak_admin.update_user(user_id=str(user_model.id), payload=payload)


def delete_business_units_to_user(user_model):
    keycloak_admin = get_default_master_keycloak_admin()
    keycloak_admin.set_realm_name(realm_name=user_model.realm_name)
    attributes = get_attributes(user_model)
    if 'business_units' in attributes:
        attributes.pop('business_units')
    payload = {
        "attributes": {
            **attributes
        }
    }
    keycloak_admin.update_user(user_id=str(user_model.id), payload=payload)


def delete_user_by_owner(user_model):
    keycloak_admin = get_default_master_keycloak_admin()
    print(user_model)
    keycloak_admin.set_realm_name(realm_name=user_model.realm_name)
    keycloak_admin.delete_user(user_id=user_model.id)


from django_keycloak.models import Realm, Client
from django_keycloak.keycloak import KeycloakAdmin
from django.contrib.auth.models import Permission, Group
from typing import List, Dict


class RolePermissionKeyclock:

    def __init__(self, realm: Realm):
        clients = Client.objects.filter(realm=realm).all()

        clients_ids = [client.client_id for client in clients]

        self.keycloak_admin = KeycloakAdmin(server_url=realm.url,
                                            username=realm.username_or_email,
                                            password=realm.password,
                                            realm_name=realm.name,
                                            client_id=realm.client_id,
                                            client_secret_key=realm.secret,
                                            auto_refresh_token=['get', 'put', 'post', 'delete'],
                                            verify=True)
        self.clients_ids = [self.keycloak_admin.get_client_id(client_id) for client_id in clients_ids]

        self.users = self.keycloak_admin.get_users({})

        self.permissions = Permission.objects.all()

        self.roles = Group.objects.all()

    def _create_permissions(self, permissions_names: List[str], client_roles: List[Dict], client_role_id: str) -> None:
        perm_names = [role['name'] for role in client_roles if role['name'].endswith('_permission')]
        create_permissions = [perm for perm in permissions_names if perm not in perm_names]
        for roleName in create_permissions:
            self.keycloak_admin.create_client_role(client_role_id=client_role_id,
                                                   payload={'name': roleName, 'clientRole': True})

    def _create_roles(self, client_roles: List[Dict], client_role_id: str) -> None:
        roles_info = []
        for rol in self.roles:
            permissions_names = ['%s_%s_%s' % (perm.content_type.app_label, perm.codename, 'permission') for perm in
                                 rol.permissions.all()]
            permissions_include = [dict(rol) for rol in client_roles if rol['name'] in permissions_names]
            roles_info.append({
                'name': '%s_%s' % (rol.name, 'role'),
                'permissions': permissions_include
            })
        roles_names = [role['name'] for role in client_roles if role['name'].endswith('_role')]
        create_roles = [rol for rol in roles_info if rol['name'] not in roles_names]
        update_roles = [rol for rol in roles_info if rol['name'] in roles_names]
        for rol in create_roles:
            self.keycloak_admin.create_client_role(client_role_id=client_role_id,
                                                   payload={'name': rol['name'], 'clientRole': True})
            self.keycloak_admin.add_composite_client_roles_to_role(client_role_id=client_role_id, role_name=rol['name'],
                                                                   roles=rol['permissions'])

        for rol in update_roles:
            roles_info = self.keycloak_admin.get_composite_client_roles_to_role(client_role_id=client_role_id,
                                                                                role_name=rol['name'])
            permissions_assign = [permission for permission in rol['permissions'] if
                                  permission['name'] not in [r['name'] for r in roles_info]]
            permissions_remove = [r for r in roles_info if
                                  r['name'] not in [permission['name'] for permission in rol['permissions']]]
            if permissions_assign:
                self.keycloak_admin.add_composite_client_roles_to_role(client_role_id=client_role_id,
                                                                       role_name=rol['name'], roles=permissions_assign)
            if permissions_remove:
                self.keycloak_admin.remove_composite_client_roles_to_role(client_role_id=client_role_id,
                                                                          role_name=rol['name'],
                                                                          roles=permissions_remove)

    def synchronize_permissions_roles(self):
        for client_id in self.clients_ids:
            client_roles = self.keycloak_admin.get_client_roles(client_id=client_id)
            permissions_names = ['%s_%s_%s' % (perm.content_type.app_label, perm.codename, 'permission') for perm in
                                 self.permissions]
            self._create_permissions(permissions_names=permissions_names, client_roles=client_roles,
                                     client_role_id=client_id)
            client_roles = self.keycloak_admin.get_client_roles(client_id=client_id)
            self._create_roles(client_roles=client_roles, client_role_id=client_id)



import unicodedata
import uuid
from typing import List
from django.contrib.auth import password_validation
from django.contrib.auth.base_user import BaseUserManager
from django.contrib.auth.hashers import (
    check_password,
    is_password_usable,
    make_password,
)
from django.contrib.auth.validators import UnicodeUsernameValidator
from django.core.mail import send_mail
from django.utils import timezone
from django.contrib.auth.models import Permission, Group
from django.db import models
from django.utils.crypto import salted_hmac
from django.utils.translation import gettext_lazy as _
from django.apps import apps
from django.db import transaction
from .services.user_repository import create_user_by_owner
from django.contrib.auth import get_user_model


class UserManager(BaseUserManager):
    use_in_migrations = True

    def _create_user(self, username, email, password, **extra_fields):
        """
        Create and save a user with the given username, email, and password.
        """
        if not username:
            raise ValueError("The given username must be set")
        with transaction.atomic():
            email = self.normalize_email(email)
            # Lookup the real model class from the global app registry so this
            # manager method can be used in migrations. This is fine because
            # managers are by definition working on the real model.
            GlobalUserModel = apps.get_model(
                self.model._meta.app_label, self.model._meta.object_name
            )
            username = GlobalUserModel.normalize_username(username)
            user = self.model(username=username, email=email, **extra_fields)
            user.password = make_password(password)
            user.save(using=self._db)
            user_id = create_user_by_owner(user, password=password)
            get_user_model().objects.filter(id=user.id).update(id=user_id)
            user.id = user_id
            return user


    def create_user(self, username, email=None, password=None, **extra_fields):
        extra_fields.setdefault("is_staff", False)
        extra_fields.setdefault("is_superuser", False)
        return self._create_user(username, email, password, **extra_fields)

    def create_superuser(self, username, email=None, password=None, **extra_fields):
        extra_fields.setdefault("is_staff", True)
        extra_fields.setdefault("is_superuser", True)
        extra_fields.setdefault("email_verified", True)

        if extra_fields.get("is_staff") is not True:
            raise ValueError("Superuser must have is_staff=True.")
        if extra_fields.get("is_superuser") is not True:
            raise ValueError("Superuser must have is_superuser=True.")
        if extra_fields.get("email_verified") is not True:
            raise ValueError("Superuser must have email_verified=True.")

        return self._create_user(username, email, password, **extra_fields)

    def with_perm(
        self, perm, is_active=True, include_superusers=True, backend=None, obj=None
    ):
        return self.none()


class keycloakUserAbstract(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    password = models.CharField(_("password"), max_length=128)
    # Stores the raw password if set_password() is called so that it can
    # be passed to password_changed() after the model is saved.
    _password = None
    last_login = models.DateTimeField(_("last login"), blank=True, null=True)
    email_verified = models.BooleanField(
        _("email status"),
        default=False
    )

    """
        An abstract base class implementing a fully featured User model with
        admin-compliant permissions.

        Username and password are required. Other fields are optional.
        """

    username_validator = UnicodeUsernameValidator()

    username = models.CharField(
        _("username"),
        max_length=150,
        unique=True,
        help_text=_(
            "Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only."
        ),
        validators=[username_validator],
        error_messages={
            "unique": _("A user with that username already exists."),
        },
    )
    first_name = models.CharField(_("first name"), max_length=150, blank=True)
    last_name = models.CharField(_("last name"), max_length=150, blank=True)
    email = models.EmailField(_("email address"), blank=True, unique=True)
    is_staff = models.BooleanField(
        _("staff status"),
        default=False,
        help_text=_("Designates whether the user can log into this admin site."),
    )
    is_active = models.BooleanField(
        _("active"),
        default=True,
        help_text=_(
            "Designates whether this user should be treated as active. "
            "Unselect this instead of deleting accounts."
        ),
    )
    date_joined = models.DateTimeField(_("date joined"), default=timezone.now)

    objects = UserManager()

    EMAIL_FIELD = "email"
    USERNAME_FIELD = 'email'
    REQUIRED_FIELDS = ['username', 'first_name', 'last_name']

    """
         Add the fields and methods necessary to support the Group and Permission
         models using the ModelBackend.
         """

    is_superuser = models.BooleanField(
        _("superuser status"),
        default=False,
        help_text=_(
            "Designates that this user has all permissions without "
            "explicitly assigning them."
        ),
    )
    permissions: List[Permission] = []
    groups: List[Group] = []

    class Meta:
        verbose_name = _("user")
        verbose_name_plural = _("users")
        abstract = True

    def get_username(self):
        """Return the username for this User."""
        return getattr(self, self.USERNAME_FIELD)

    def clean(self):
        setattr(self, self.USERNAME_FIELD, self.normalize_username(self.get_username()))
        self.email = self.__class__.objects.normalize_email(self.email)


    def natural_key(self):
        return self.get_username(),


    @property
    def is_anonymous(self):
        """
        Always return False. This is a way of comparing User objects to
        anonymous users.
        """
        return False

    @property
    def is_authenticated(self):
        """
        Always return True. This is a way to tell if the user has been
        authenticated in templates.
        """
        return True

    def set_password(self, raw_password):
        self.password = make_password(raw_password)
        self._password = raw_password

    def check_password(self, raw_password):
        """
        Return a boolean of whether the raw_password was correct. Handles
        hashing formats behind the scenes.
        """

        def setter(raw_password):
            self.set_password(raw_password)
            # Password hash upgrades shouldn't be considered password changes.
            self._password = None
            self.save(update_fields=["password"])

        return check_password(raw_password, self.password, setter)

    def set_unusable_password(self):
        # Set a value that will never be a valid hash
        self.password = make_password(None)

    def has_usable_password(self):
        """
        Return False if set_unusable_password() has been called for this user.
        """
        return is_password_usable(self.password)

    def get_session_auth_hash(self):
        """
        Return an HMAC of the password field.
        """
        key_salt = "django.contrib.auth.models.AbstractBaseUser.get_session_auth_hash"
        return salted_hmac(
            key_salt,
            self.password,
            algorithm="sha256",
        ).hexdigest()

    @classmethod
    def get_email_field_name(cls):
        try:
            return cls.EMAIL_FIELD
        except AttributeError:
            return "email"

    @classmethod
    def normalize_username(cls, username):
        return (
            unicodedata.normalize("NFKC", username)
            if isinstance(username, str)
            else username
        )

    def get_full_name(self):
        """
        Return the first_name plus the last_name, with a space in between.
        """
        full_name = "%s %s" % (self.first_name, self.last_name)
        return full_name.strip()

    def get_short_name(self):
        """Return the short name for the user."""
        return self.first_name

    def email_user(self, subject, message, from_email=None, **kwargs):
        """Send an email to this user."""
        send_mail(subject, message, from_email, [self.email], **kwargs)

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        if self._password is not None:
            password_validation.password_changed(self._password, self)
            self._password = None


    def get_user_permissions(self, obj=None):
        """
        Return a list of permission strings that this user has directly.
        Query all available auth_panel backends. If an object is passed in,
        return only permissions matching this object.
        """
        return set(self.permissions)

    def get_group_permissions(self, obj=None):
        """
        Return a list of permission strings that this user has through their
        groups. Query all available auth_panel backends. If an object is passed in,
        return only permissions matching this object.
        """
        return set(self.permissions)

    def get_all_permissions(self, obj=None):
        return set([perm.name for perm in self.permissions])

    def has_perm(self, perm, obj=None):
        """
        Return True if the user has the specified permission. Query all
        available auth_panel backends, but return immediately if any backend returns
        True. Thus, a user who has permission from a single auth_panel backend is
        assumed to have permission in general. If an object is provided, check
        permissions for that object.
        """

        if self.is_active and self.is_superuser:
            return True

        # Otherwise we need to check the backends.
        return perm in self.get_all_permissions()

    def has_perms(self, perm_list, obj=None):
        """
        Return True if the user has each of the specified permissions. If
        object is passed, check if the user has all required perms for it.
        """
        if self.is_active and self.is_superuser:
            return True

        return all(self.has_perm(perm, obj) for perm in perm_list)

    def has_module_perms(self, app_label):
        """
        Return True if the user has any permissions in the given app label.
        Use similar logic as has_perm(), above.
        """
        if self.is_active and self.is_superuser:
            return True

        # Active superusers have all permissions.
        return self.is_active and any(
            perm[:perm.index('.')] == app_label
            for perm in self.get_all_permissions()
        )

    def __str__(self) -> str:
        return str(self.id)


class User(keycloakUserAbstract):
    first_name = models.CharField(_('first name'), max_length=150, blank=True)
    last_name = models.CharField(_('last name'), max_length=150, blank=True)

    class Meta:
        managed = False
        db_table = 'user_keycload'
        verbose_name = 'User'
        verbose_name_plural = 'Users'


class Realm(models.Model):
    url = models.CharField(max_length=255, null=False, blank=False,
                            help_text='Url as known on the Keycloak server. '
                                      'This url is used in the API paths '
                                      'of this Realm.')
    name = models.CharField(max_length=255, unique=True, null=False, blank=False,
                            help_text='Name as known on the Keycloak server. '
                                      'This name is used in the API paths '
                                      'of this Realm.')
    client_id = models.CharField(max_length=255, null=False, blank=False,
                                 help_text='Client id as known on the Keycloak server. '
                                      'This name is used in the API paths '
                                      'of this Realm.')
    secret = models.CharField(max_length=255, null=False, blank=False,
                              help_text='Secret of client id on the Keycloak server.')
    username_or_email = models.CharField(max_length=255, null=False, blank=False,
                              help_text='User of client id on the Keycloak server.')
    password = models.CharField(max_length=255, null=False, blank=False,
                              help_text='Password of user on the Keycloak server.')

    def __str__(self):
        return self.name


    class Meta:
        managed = True
        db_table = 'realm'
        verbose_name = 'Realm'
        verbose_name_plural = 'Realms'

        
class Client(models.Model):
    realm = models.ForeignKey(Realm, related_name='client',
                                 on_delete=models.CASCADE)
    client_id = models.CharField(max_length=255)
    

    def __str__(self):
        return self.client_id

    class Meta:
        managed = True
        db_table = 'realm_client'
        verbose_name = 'Realm Client'
        verbose_name_plural = 'Realm Clients'




from django_keycloak.keycloak import get_default_master_keycloak_admin


def create_realm_by_owner(realm_model):
    keycloak_admin = get_default_master_keycloak_admin()
    payload = dict(realm_model.base_config)
    payload['id'] = str(realm_model.id)
    payload['realm'] = realm_model.realm
    if realm_model.email_realm_config:
        payload['smtpServer'] = {
            "replyToDisplayName": realm_model.email_realm_config.replyToDisplayName,
            "starttls": realm_model.email_realm_config.starttls,
            "auth": realm_model.email_realm_config.auth,
            "envelopeFrom": realm_model.email_realm_config.envelopeFrom,
            "ssl": realm_model.email_realm_config.ssl,
            "password": realm_model.email_realm_config.password,
            "port": realm_model.email_realm_config.port,
            "replyTo": realm_model.email_realm_config.replyTo,
            "host": realm_model.email_realm_config.host,
            "from": realm_model.email_realm_config.from_email,
            "fromDisplayName": realm_model.email_realm_config.fromDisplayName,
            "user": realm_model.email_realm_config.user
        }
    else:
        payload.pop('smtpServer')

    keycloak_admin.create_realm(payload=payload)


def update_realm_by_owner(realm_model):
    keycloak_admin = get_default_master_keycloak_admin()
    payload = dict(realm_model.base_config)
    payload.pop('id')
    payload.pop('realm')
    payload['smtpServer'] = {}
    keycloak_admin.update_realm(realm_name=realm_model.realm, payload=payload)
    if realm_model.email_realm_config:
        payload['smtpServer'] = {
            "replyToDisplayName": realm_model.email_realm_config.replyToDisplayName,
            "starttls": realm_model.email_realm_config.starttls,
            "auth": realm_model.email_realm_config.auth,
            "envelopeFrom": realm_model.email_realm_config.envelopeFrom,
            "ssl": realm_model.email_realm_config.ssl,
            "password": realm_model.email_realm_config.password,
            "port": realm_model.email_realm_config.port,
            "replyTo": realm_model.email_realm_config.replyTo,
            "host": realm_model.email_realm_config.host,
            "from": realm_model.email_realm_config.from_email,
            "fromDisplayName": realm_model.email_realm_config.fromDisplayName,
            "user": realm_model.email_realm_config.user
        }
        keycloak_admin.update_realm(realm_name=realm_model.realm, payload=payload)


def delete_realm_by_owner(realm_model):
    keycloak_admin = get_default_master_keycloak_admin()
    keycloak_admin.delete_realm(realm_name=realm_model.realm)


from django_keycloak.keycloak import get_default_master_keycloak_admin


def get_permissions_by_app(apps_owners_model):
    keycloak_admin = get_default_master_keycloak_admin()
    keycloak_admin.set_realm_name(realm_name=apps_owners_model.owner.name)
    roles = keycloak_admin.get_client_roles(client_id=apps_owners_model.id)
    return [p for p in roles if p['name'].endswith('_permission')]


def create_permission_by_app(earp_model):
    keycloak_admin = get_default_master_keycloak_admin()
    keycloak_admin.set_realm_name(
        realm_name=earp_model.app_role.apps_owner.owner.name)
    permissions = get_permissions_by_app(earp_model.app_role.apps_owner)
    permissions_names = [p['name'] for p in permissions]
    name = earp_model.endpoint.name
    name = name.replace(' ', '_').lower()
    all_permissions = [f'{name}_allow_read_permission',
                       f'{name}_allow_add_permission',
                       f'{name}_allow_delete_permission',
                       f'{name}_allow_update_permission',
                       f'{name}_allow_execute_permission']
    permissions_create = [
        p for p in all_permissions if p not in permissions_names]
    for p in permissions_create:
        payload = {
            "name": p,
            "description": p.replace("_", " "),
            "clientRole": True,
            "attributes": {
                "is_editable": [False],
                "is_role": [False]
            }
        }
        client_role_id = keycloak_admin.get_client_id(
            client_id=earp_model.app_role.apps_owner.app.client_id)
        keycloak_admin.create_client_role(
            client_role_id=client_role_id, payload=payload)


def add_permission_by_role_app(earp_model):
    keycloak_admin = get_default_master_keycloak_admin()
    keycloak_admin.set_realm_name(
        realm_name=earp_model.app_role.apps_owner.owner.name)
    permissions = get_permissions_by_app(earp_model.app_role.apps_owner)
    name = earp_model.endpoint.name
    name = name.replace(' ', '_').lower()
    all_permissions = []
    if earp_model.allow_read:
        all_permissions.append(f'{name}_allow_read_permission')
    if earp_model.allow_add:
        all_permissions.append(f'{name}_allow_add_permission')
    if earp_model.allow_delete:
        all_permissions.append(f'{name}_allow_delete_permission')
    if earp_model.allow_update:
        all_permissions.append(f'{name}_allow_update_permission')
    if earp_model.allow_execute:
        all_permissions.append(f'{name}_allow_execute_permission')

    permissions_append = [
        p for p in permissions if p['name'] in all_permissions]
    client_role_id = keycloak_admin.get_client_id(
        client_id=earp_model.app_role.apps_owner.app.client_id)
    role_name = earp_model.app_role.role.name.replace(' ', '_').lower()
    delete_roles = keycloak_admin.get_composite_client_roles_to_role(
        client_role_id=client_role_id, role_name=f'{role_name}_role')
    delete_roles = [r for r in delete_roles if r['name'].startswith(f'{name}_allow')]
    if delete_roles:
        keycloak_admin.remove_composite_client_roles_to_role(
            client_role_id=client_role_id, role_name=f'{role_name}_role', roles=delete_roles)
    keycloak_admin.add_composite_client_roles_to_role(client_role_id=client_role_id, role_name=f'{role_name}_role', roles=permissions_append)


def delete_permission_by_app(earp_model):
    keycloak_admin = get_default_master_keycloak_admin()
    keycloak_admin.set_realm_name(
        realm_name=earp_model.app_role.apps_owner.owner.name)
    permissions = get_permissions_by_app(earp_model.app_role.apps_owner)
    permissions_names = [p['name'] for p in permissions]
    name = earp_model.endpoint.name
    all_permissions = [f'{name}_allow_read_permission',
                       f'{name}_allow_add_permission',
                       f'{name}_allow_delete_permission',
                       f'{name}_allow_update_permission',
                       f'{name}_allow_execute_permission']
    permissions_delete = [
        p for p in permissions_names if p in all_permissions]
    client_role_id = keycloak_admin.get_client_id(
        client_id=earp_model.app.client_id)
    for p in permissions_delete:
        keycloak_admin.delete_client_role(client_role_id=client_role_id, role_name=p)

